#!/usr/bin/perl
#
# Written by Michael C. Richardson on 1996-11-29
# Copyright Sandelman Software Works. http://www.sandelman.ottawa.on.ca/
#
# You can do whatever you like with this code so long as you give me credit.
#
# $Id: build_virtual_mail.pl,v 1.4 1997/05/15 14:31:53 mcr Exp mcr $
# 
# $Log: build_virtual_mail.pl,v $
# Revision 1.4  1997/05/15 14:31:53  mcr
# Virtual POP additions.
#
# Revision 1.3  1997/05/14 19:13:08  mcr
# added virtual host to makefile.
#
# Revision 1.2  1996/12/03 20:53:12  mcr
# Working copy.
#
# Revision 1.1  1996/11/30 22:20:53  mcr
# Initial revision
#
#

require 'getopts.pl';

# =()<$top="@<libdir>@";>()=
$configdir="/etc/virtualmail";
$configfile="$configdir/config";
$spooldir="/var/virtualmail";

&Getopts('C:d');

if($opt_C) {
    $configfile=$opt_C;
}

if($opt_d) {
    $debug=1;
}

open(CONFIG,"$configfile") || die "Can not open $configfile: $!\n";
while(<CONFIG>) {
    next if (/^\#/);
    chop;
    if(/^virtualmailprefix[ \t]+(.*)/) {
	$mailprefix=$1;
    }
    if(/^configdir[ \t]+(.*)/) {
	$configdir=$1;
    }
    if(/^spooldir[ \t]+(.*)/) {
	$spooldir=$1;
    }
    if(/^rootdir[ \t]+(.*)/) {
	$rootdir=$1;
    }
}
close(CONFIG);

print "Build Virtual mail: $mailprefix, \n\t$configdir, \n\t$spooldir\n" if($debug);

dbmopen(%popusers, "$configdir/popusers", 0600) || die "$configdir/popusers: $!\n";

$activedomains="$configdir/activedomains";

chop($date=`date`);

open(ACTIVEDOMAIN,"$activedomains");
while(<ACTIVEDOMAIN>) {
    next if (/^\#/);
    chop;
    $domain   = $_;

    unshift(@domains,$domain);

    if(!$processthings) {
	
	$mailconf="$mailprefix/$domain/mailconf";
	opendir(CONFDIR,$mailconf);
	@files=readdir(CONFDIR);
	closedir(CONFDIR);
	
	foreach $file (@files) {
	    if(&newer("$mailconf/$file","$configdir/virtualaliases")) {
		
		print "Processing due to new $mailconf/$file\n";
		$processthings=1;
		last;
	    }
	}
    }
}
close(ACTIVEDOMAIN);

# anything to do?
exit unless ($processthings);


foreach $domain (@domains) {
    print "Processing directory $domain\n" if ($debug);
    &processdomaindir($domain);
}

open(ALIASFILE,">$configdir/virtualaliases");
print ALIASFILE "# this file was generated on $date\n";
print ALIASFILE join("\n",@aliases)."\n";
close(ALIASFILE);
system("/usr/ucb/newaliases");

open(MAILBASE,">$configdir/mailbase");
print MAILBASE "# this file was generated on $date\n";
print MAILBASE join("\n",@mailbase)."\n";
close(MAILBASE);
system("/usr/sbin/makemap -r hash $configdir/mailbase.db <$configdir/mailbase");

# generate the virtual root file
open(VIRTUALROOT,">/etc/virtualmail/roots");
print VIRTUALROOT "# this file was generated on $date\n";
foreach $domain (keys %domainpop) {
    $ip=&ipforhost("www.$domain");
    print VIRTUALROOT "$domain $ip $spooldir/$domain/root\n";
}
close(VIRTUALROOT);    

# now, regenerate password files
foreach $domain (keys %domaintouched) {
    system("/etc/virtualmail/virthost -m $domain /usr/sbin/pwd_mkdb /etc/master.passwd");
}

dbmclose(%popusers);

if($errors) {
    print "There were $errors errors\n";
    exit 1;
}

# END

sub ipforhost {
    local($host) = @_;

    ($name,$aliases,$addrtype,$length,@addrs)=gethostbyname($host);
    $ip=$addrs[0];
    ($a,$b,$c,$d)=unpack('C4',$ip);
    return "$a.$b.$c.$d";
}

sub sanifydestaddr {
    local($addr)=@_;

    # check that this is a real email address
    if($addr =~ m,\@\,\<\>\",) {
	print STDERR "File $redirector:$line has bad destination address $addr\n";
	$errors++;
	return undef;
    }
    return $addr;
}

sub sanifylocaladdr {
    local($addr)=@_;
    
    if($addr =~ m,\@\,\<\>\",) {
	print STDERR "File $redirector:$line has bad local address $addr\n";
	$errors++;
	return undef;
    }
    return $addr;
}

sub forwards {
    local($from,$to)=@_;
    
#    print "$from -> $to\n";
    unshift(@mailbase,"$from\t$to");
}

sub popuser {
    local($from,$domain)=@_;
    local($uid,$rootdir);
    
    &forwards("$from\@$domain","=POPMAIL $from\@$domain");

    $domainpop{$domain}++;

    if(!$popusers{"$from\@$domain"}) {
	$popusers{"$from\@$domain"}=1;

	$uid=$popusers{"highest"}++;
	$uid+=10000;

	$rootdir="$spooldir/$domain/root";
	open(PASSWDFILE,">>$rootdir/etc/master.passwd");
#                         daemon:*:1:31::0:0:The devil himself:/root:

	print PASSWDFILE "$from::$uid:32766::0:0:A user:/none:/none\n";
	close(PASSWDFILE);
	
	$domaintouched{$domain}++;
    }
}

sub newer {
    local($one,$two)=@_;

    $mtime1=(stat($one))[9];
    $mtime2=(stat($two))[9];
    if($mtime1 > $mtime2) {
	return 1;
    } else {
	return 0;
    }
}

#
# this routine is here for a couple of reasons
#    1. you can not mail directly to an :include: file
#    2. the ownership on the :include: file must be right
#    3. the contents of the :include: file should be checked out.
#
sub explodeto {
    local($domain,$domaindir,$origfile)=@_;

    local($newname,$names,@names,$alias);
   
    system("mkdir $spooldir/$domain") unless(-d "$spooldir/$domain");

    $newname="$spooldir/$domain/$origfile";

    $alias="$domain-$origfile";

    if(&newer("$domaindir/$origfile",$newname)) {
	# got to process things
	open(NEWFILE,">$newname");
	open(OLDFILE,"$domaindir/$origfile");
	@names=<OLDFILE>;
	
	#regularize input
	$names=join(",",@names);
	$names=~ y/\n//d;	# nuke all newlines
	@names=split(/,/,$names);
	
	foreach (@names) {
	    if((/^\// && !/\@/) || /^\|/ || /^:include:/ ) {
		print STDERR "Exploder checking $domain, $origfile bad address $_\n";
		$errors++;
		next;
	    } 
	    print NEWFILE "$_\n";
	}
    }

    unshift(@aliases,"$alias: :include:$newname");
    return "$alias";
}

#
# this routine is here for a couple of reasons
#    1. you can not mail directly to a program
#    2. the ownership is not right anyway.
#    3. you can not expand to more than one name
#
sub progto {
    local($domain,$domaindir,$responsefile,$notifyname)=@_;

    local($newname,$names,@names,$alias);
   
    $num++;
    $alias=sprintf("%s-resp%02d",$domain,$num);

    if(defined($notifyname) && $notifyname ne "@@") {
	unshift(@aliases,"$alias: \"|$configdir/wrapper autorespond $from $thisdomain $domaindir/$file\",$notifyname");
    } else {
	unshift(@aliases,"$alias: \"|$configdir/wrapper autorespond $from $thisdomain $domaindir/$file\"");
    }	
    return "$alias";
}


sub processdomaindir {
    local($domain)=@_;
    local($domaindir,$thisdomain); 

    $thisdomain=0;
    $domaindir="$mailprefix/$domain/mailconf";
    $redirector="$domaindir/redirect";

    # check to see if 

    open(REDIRECT,"$redirector") || warn "Can not open $redirector: $!\n";
    $line=0;
    while(<REDIRECT>) {
	$line++;
	next if (/^\#/);
	chop;

	($cmd,@args)=split;
	$cmd =~ y/A-Z/a-z/;
	if(!$thisdomain) {
	    if($cmd eq "domain") {
		$thisdomain=$args[0];
	    } else {
		print STDERR "File $redirector: must start with <domain> statement\n";
		$errors++;
 	    }
	}
	else {
	    if($cmd eq "domain") {
		print STDERR "File $redirector:$line has more than one <domain> statement\n";
		$errors++;
	    } elsif($cmd eq "forward") {
		$from=&sanifylocaladdr($args[0]);
		$to=&sanifydestaddr($args[1]);
		next unless (defined($from) && defined($to));
		&forwards("$from\@$thisdomain",$to);

	    } elsif($cmd eq "popuser") {
		$from=&sanifylocaladdr($args[0]);
		next unless (defined($from));
		&popuser($from,$thisdomain);

	    } elsif($cmd eq "responder") {
		$from=&sanifylocaladdr($args[0]);
		next unless(defined($from));

		$file=$args[1];
		if($file =~ m,\W,) {
		    print STDERR "File $redirector:$line invalid autoresponse file: $file\n";
		    $errors++;
		    next;
		}
		if(!-f "$domaindir/$file") {
		    print STDERR "File $redirector:$line asked for non-existant response file $file\n";
		    $errors++;
		}
		$notify=$args[2];
		$progname=&progto($thisdomain,$domaindir,$file,$notify);
		&forwards("$from\@$thisdomain",$progname);

	    } elsif($cmd eq "exploder") {
		$from=&sanifylocaladdr($args[0]);
		next unless(defined($from));

		$file=$args[1];
		if($file =~ m,\W,) {
		    print STDERR "File $redirector:$line invalid exploder file: $file\n";
		    $errors++;
		    next;
		}
		if(!-f "$domaindir/$file") {
		    print STDERR "File $redirector:$line asked for non-existant exploder file $file\n";
		    $errors++;
		}
		$exploder_name=&explodeto($thisdomain,$domaindir,$file);
		&forwards("$from\@$thisdomain",$exploder_name);
		
	    } elsif($cmd eq "bounce") {
		&forwards($thisdomain,"=ERROR ".join(' ',@args));
	    } elsif($cmd eq "smartuser") {
		&forwards($thisdomain,$args[0]);
	    } else {
		printf STDERR "File $redirector:$line invalid command $cmd\n";
		$errors++;
	    }
	}
    }
    close(REDIRECT);
}

	    



