/*
 * Run a command in a chrooted environment
 *
 * The choice of where to chroot to is determined by examining
 * the socket on stdin. The local (getsockname(2)) address is used
 * to lookup the directory to chroot to from the file
 * /etc/virtualmail/roots
 */

/*
 * Copyright (c) 1988, 1993
 *	The Regents of the University of California.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by the University of
 *	California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ifndef lint
static char copyright[] =
"@(#) Copyright (c) 1988, 1993\n\
	The Regents of the University of California.  All rights reserved.\n";
#endif /* not lint */

#ifndef lint
/* static char sccsid[] = "@(#)chroot.c	8.1 (Berkeley) 6/9/93"; */
static char rcsid[] = "$NetBSD: chroot.c,v 1.5 1997/03/06 07:00:52 mikel Exp $";
#endif /* not lint */

#include <sys/types.h>

#include <err.h>
#include <errno.h>
#include <paths.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <syslog.h>
#include <sys/socket.h>
#include <sys/syslog.h>
#include <netinet/in.h>
#include <sys/un.h>
#include <signal.h>

#define _PATH_ROOTS "/etc/virtualmail/roots"

void usage __P((void));

char  buf[1024];

char *
findroot(host)
     char *host;
{
  struct sockaddr_in sin;
  struct in_addr addr;
  int   sinlen;
  char  *l,*name, *ip, *root;
  FILE  *rootfile;
  int   linenum;

  if(!host)
    {
      sinlen=sizeof(sin);
      if(getsockname(0, (struct sockaddr *)&sin, &sinlen) == -1)
	{
	  syslog(LOG_ERR,"getsockname: %m");
	  exit(1);
	}
    }

  if((rootfile=fopen(_PATH_ROOTS,"r")) == NULL)
    {
      syslog(LOG_ERR,"%s: %m", _PATH_ROOTS);
      exit(1);
    }

  linenum=0;
  while(fgets(buf, 1024, rootfile) != NULL)
    {
      linenum++;
      /*syslog(LOG_INFO,"processing %s",buf);*/
      if(buf[0]=='#') continue;

      l=buf;
      while ((name = strsep(&l, " \t\n")) != NULL && *name == '\0');
      while ((ip   = strsep(&l, " \t\n")) != NULL && *ip   == '\0');
      while ((root = strsep(&l, " \t\n")) != NULL && *root == '\0');

      /*      syslog(LOG_INFO,"name: %s ip: %s root: %s\n",name,ip,root); */
      if(host)
	{
	  if(strcasecmp(host, name) == 0)
	    {
	      syslog(LOG_INFO,"Hostname %s [%s] chrooted to %s",
		     name, ip, root);
	      return root;
	    }
	}
      else
	{
	  if(inet_aton(ip, &addr) == 1)
	    {
	      if(addr.s_addr == sin.sin_addr.s_addr)
		{
		  syslog(LOG_INFO,"Connect to %s [%s] chrooted to %s",
			 name, inet_ntoa(sin.sin_addr), root);
		  fclose(rootfile);
		  return root;
		}
	    }
	  else
	    {
	      syslog(LOG_ERR,"Bad ip address %s (line %d)",ip, linenum);
	    }
	}
    }
  
  fclose(rootfile);
  
  if(!host)
    {
      syslog(LOG_WARNING,"Connection to ip %s not chrooted", inet_ntoa(sin.sin_addr));
    }
  else
    {
      syslog(LOG_WARNING,"Host %s is not chrooted", host);
    }
  return NULL;
}


int syslogsock;

void 
setup_syslog(root)
     char *root;
{
  char buf[1024];
  struct sockaddr_un sunaddr;

  buf[0]='\0';
  strncat(buf, root, sizeof(buf));
  strncat(buf, _PATH_LOG, sizeof(buf));

  if((syslogsock = socket(PF_UNIX, SOCK_DGRAM, 0)) == -1 )
    {
      syslog(LOG_ERR,"setup_syslog socket: %m");
      exit(1);
    }
  
  unlink(buf);
  
  sunaddr.sun_path[0]='\0';
  strncat(sunaddr.sun_path, buf, sizeof(sunaddr.sun_path));
  sunaddr.sun_family=AF_UNIX;
  sunaddr.sun_len=SUN_LEN(&sunaddr);

  if(bind(syslogsock, (struct sockaddr *)&sunaddr, sunaddr.sun_len) == -1)
    {
      syslog(LOG_ERR,"setup_syslog bind (%s): %m", sunaddr.sun_path);
      exit(1);
    }

}

void
proc_syslog()
{
  char buf[1024];
  int sock, len;
  struct sockaddr_un sunaddr;

  strcpy(sunaddr.sun_path, _PATH_LOG);
  sunaddr.sun_family=AF_UNIX;
  sunaddr.sun_len=SUN_LEN(&sunaddr);

  if((sock = socket(PF_UNIX, SOCK_DGRAM, 0)) == -1 )
    {
      syslog(LOG_ERR,"proc_syslog socket: %m");
      exit(1);
    }

  if(connect(sock, (struct sockaddr *)&sunaddr, sunaddr.sun_len) == -1)
    {
      syslog(LOG_ERR,"proc_syslog connect: %m");
      exit(1);
    }

  while((len = read(syslogsock, buf, 1024)) > 0)
    {
      /*      syslog(LOG_INFO,"read %d bytes\n",len);*/
      if(write(sock, buf, len) != len)
	{
	  syslog(LOG_ERR,"proc_syslog write: %m");
	  exit(1);
	}
    }

  close(sock);
  close(syslogsock);
  exit(0);
}

void
virthost_exit(sig, code, scp)
     int sig, code;
     struct sigcontext *scp;
{
  int stat;

  (void)wait(&stat);
  syslog(LOG_INFO,"exiting with %d\n",stat);
  exit(0);
}

int
main(argc, argv)
     int argc;
     char *argv[];
{
  int ch;
  char *shell;
  char *root;
  char *host = NULL;
  int  pid;
  struct sigaction sa1;

  openlog("virthost", LOG_PID, LOG_DAEMON);

  /* printf("argc: %d\n", argc); */

  while ((ch = getopt(argc, argv, "m:")) != -1)
    {
      switch(ch) {
      case 'm':
	host=optarg;
	break;
	      
      case '?':
      default:
	usage();
      }
    }

  argc -= optind;
  argv += optind;

  /* printf("optind: %d argc: %d path: %s\n", optind, argc, argv[0]); */
  if (argc < 0)
    usage();

  if (!argv[0]) {
    usage();
  }

  root = findroot(host);

  if(root) {
    setup_syslog(root);
	  
    pid=fork();
	  
    switch(pid) {
    case 0:
      /* child */
      if (chdir(root) || chroot("."))
	{
	  syslog(LOG_ERR, "chroot %s: %m", root);
	  exit(1);
	}
	    
      execvp(argv[0], &argv[0]);
      syslog(LOG_ERR, "exec %s: %m", argv[0]);
      exit(1);
      /* NOTREACHED */
      break;
      
    case -1:
      syslog(LOG_ERR, "fork: %m");
      exit(1);
      
    default:
      /* in parent */
      
      /* setup signal handler */
      sa1.sa_handler = virthost_exit;
      sigemptyset(&sa1.sa_mask);
      sa1.sa_flags   = 0;
      if(sigaction(SIGCHLD, &sa1, NULL) == -1)
	{
	  syslog(LOG_ERR,"sigaction: %m");
	  exit(1);
	}
      
      proc_syslog();
      exit(0);
    }
  }
  else {
    execvp(argv[0], &argv[0]);
    syslog(LOG_ERR, "exec %s: %m", argv[0]);
    exit(1);
  }	  
}

void
usage()
{
	(void)fprintf(stderr, "usage: virthost [-m host] command\n");
	exit(1);
}
